from mysticML.utils import Daddy
import pandas
import numpy as np
from sklearn.utils import resample

class Sampling(Daddy):
    def __init__(self, x:pandas.DataFrame, target:str=None,**kwargs):
        super().__init__(x, target, **kwargs)
        self.col_bool = []
        self.val = {}
        self.lth = None
        self.rth = None

    def check(self) ->bool:
        if not self.sampling:
            return False
        if self.y.nunique() > 0.0275 * self.x.shape[0] or self.y.dtype != 'object':
            return False
        for c,d in zip(self.y.value_counts().index.to_list(),self.y.value_counts()):
            self.val[c]=d
        mean = np.round(np.mean(list(self.val.values())))
        self.lth = round(mean - 0.2 * mean)
        self.rth = round(mean + 0.2 * mean)
        for i in self.val.values():
            if i<self.lth or i>self.rth:
                self.col_bool.append(True)
            else:
                self.col_bool.append(False)
        return any(self.col_bool)

    def apply(self) ->pandas.DataFrame:
        if self.check():
            self.report_.append("Sampling")
            df = pandas.DataFrame()
            c = -1
            for i in self.col_bool:
                c+=1
                if i:
                    d = self.preD(c)
                    samples = round((((list(self.val.values())[c]-min(self.val.values()))*(self.rth-self.lth))
                                                /(max(self.val.values())-min(self.val.values())))+self.lth)
                    d = resample(d, n_samples=samples)
                    df = pandas.concat([df,d])
                else:
                    df = pandas.concat([df,self.preD(c)])
            return df
        else:
            return pandas.concat([self.x, self.y],axis=1)

    def preD(self, c:int) ->pandas.DataFrame:
        d = self.x.copy()
        d[self.target] = self.y[self.y==list(self.val.keys())[c]]
        d.dropna(inplace=True)
        return d



from mysticML.utils import Daddy
import pandas
from sklearn.decomposition import PCA
from math import ceil
pandas.options.mode.chained_assignment=None

class FeatureExtraction(Daddy):

    def __init__(self,x:pandas.DataFrame, target:str=None, **kwargs)->None:
        super().__init__(x, target, **kwargs)
        self.pca = None
        self.s = None

    def check(self) ->bool:
        if not self.feature_ext:
            return False
        self.s = self.x.shape[1]
        if self.x.shape[0] < self.s:
            self.s = self.x.shape[0]
            self.pca = PCA(n_components=ceil(self.s/2))
            return True
        if self.s > 6:
            self.pca = PCA(n_components=ceil(self.s/2))
            return True
        return False

    def apply(self) ->pandas.DataFrame:
        if not self.check():
            return pandas.concat([self.x, self.y], axis=1)
        self.report_.append("FeatureExt")
        columns = ['pca_comp_%i' % i for i in range(ceil(self.s/2))]
        x = pandas.DataFrame(self.pca.fit(self.x).transform(self.x), columns=columns)
        return pandas.concat([x, self.y], axis=1)



import pandas

class Daddy:

    report_ = []

    def __init__(self, x:pandas.DataFrame, target:str=None, combo:str='advanced',
                 duplicate: bool = None, outlier: bool = None, impute: bool = None,
                 transform: bool = None, encode: bool = None,
                 feature_sel: bool = None, feature_ext: bool = None, sampling: bool = None) ->None:

        if target is not None:
            self.x = x.drop([target],axis=1)
            self.y = x[target]
            self.target = target
        else:
            self.x = x.iloc[:,:-1]
            self.y = x.iloc[:,-1]
            self.target = x.columns[-1]

        if duplicate is False:
            self.duplicate = False
        else:
            self.duplicate = True
        if outlier is False:
            self.outlier = False
        else:
            self.outlier = True
        if impute is False:
            self.impute = False
        else:
            self.impute = True
        if transform is False:
            self.transform = False
        else:
            self.transform = True
        if encode is False:
            self.encode = False
        else:
            self.encode = True
        if feature_sel is False:
            self.feature_sel = False
        else:
            self.feature_sel = True
        if feature_ext is False:
            self.feature_ext = False
        else:
            self.feature_ext = True
        if sampling is False:
            self.sampling = False
        else:
            self.sampling = True

        assert combo in ['basic', 'intermediate', 'advanced'], f"unknown value {combo} passed " \
                                                               f"for parameter combo." \
                                                               f"\n allowed values: " \
                                                               f"['basic', 'intermediate', 'advanced']"
        self.combo = combo

        if combo == 'intermediate':
            if not feature_sel:
                self.feature_sel = False
            if not feature_ext:
                self.feature_ext = False
            if not sampling:
                self.sampling = False

        elif combo == 'basic':
            if not feature_sel:
                self.feature_sel = False
            if not feature_ext:
                self.feature_ext = False
            if not sampling:
                self.sampling = False
            if not encode:
                self.encode = False
            if not transform:
                self.transform = False

        if encode is False:
            assert feature_sel is False or feature_sel is None, "feature_sel cannot be set to True when encode is set to False."
            assert feature_ext is False or feature_ext is None, "feature_ext cannot be set to True when encode is set to False."
            self.feature_sel = False
            self.feature_ext = False

        if impute is False:
            assert sampling is False or sampling is None, "sampling cannot be set to True when imputation is set to False."
            assert feature_sel is False or feature_sel is None, "feature_sel cannot be set to True when imputation is set to False."
            assert feature_ext is False or feature_ext is None, "feature_ext cannot be set to True when imputation is set to False."
            self.sampling = False
            self.feature_sel = False
            self.dim_red = False


    def check(self) ->bool:
        raise NotImplementedError
    def apply(self) ->pandas.DataFrame:
        raise NotImplementedError


import pandas
from scipy.stats import zscore
from mysticML.utils import Daddy

class Outlier(Daddy):
    def __init__(self, x: pandas.DataFrame, target:str=None, **kwargs) -> None:
        super().__init__(x, target, **kwargs)

    def check(self) -> bool:
        if not self.outlier:
            return False
        return any(self.x[col].dtype in [float, int] for col in self.x.columns)

    def apply(self) -> pandas.DataFrame:
        if not self.check():
            return pandas.concat([self.x, self.y], axis=1)
        self.report_.append("Outlier")
        for col in self.x.select_dtypes(include=[float, int]).columns:
            z_scores = zscore(self.x[col])
            outliers = (z_scores > 1.5) | (z_scores < -1.5)
            self.x.loc[outliers, col] = pandas.NA
        return pandas.concat([self.x, self.y], axis=1)


from mysticML.utils import Daddy
import pandas
import numpy as np
from sklearn.feature_selection import VarianceThreshold
from sklearn.feature_selection import SelectKBest, f_classif, f_regression
pandas.options.mode.chained_assignment=None

class FeatureSelection(Daddy):
    def __init__(self, x:pandas.DataFrame, target:str=None,**kwargs)->None:
        super().__init__(x, target, **kwargs)
        self.app = None

    def check(self) ->bool:
        if not self.feature_sel:
            return False
        var = VarianceThreshold(threshold=0.0)
        try:
            var.fit(self.x)
            low_col = [column for column in self.x.columns
                       if column not in self.x.columns[var.get_support()]]
            self.x.drop(low_col, axis=1, inplace=True)
        except ValueError:
            # print('raised error by var.fit')
            pass
        s = self.x.shape[1]
        if s > 6:
            if self.y.nunique() > 0.0275*self.x.shape[0] and self.y.dtype != 'object':
                self.app = SelectKBest(score_func=f_regression, k=int(s/2))
            else:
                self.app = SelectKBest(score_func=f_classif, k=int(s/2))
            return True
        else:
            return False

    def apply(self) ->pandas.DataFrame:
        if not self.check():
            return pandas.concat([self.x,self.y],axis=1)
        # print(self.x.nunique())
        self.report_.append("FeatureSel")
        scores = self.app.fit(self.x,self.y).scores_
        average = np.average(scores)
        values = []
        compare = self.app.get_support(indices=True)
        for i in scores:
            if i>average:
                values.append(True)
            else:
                values.append(False)
        if values.count(True)<len(compare):
            self.x = self.x.iloc[:,compare]
        else:
            self.x = self.x.iloc[:,values]
        return pandas.concat([self.x, self.y], axis=1)



import pandas
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from scipy.stats import shapiro
from scipy.stats import kstest, norm
from mysticML.utils import Daddy
import warnings

class Transformation(Daddy):
    def __init__(self, x: pandas.DataFrame,target:str=None,**kwargs) ->None:
        super().__init__(x, target, **kwargs)

    def check(self) ->bool:
        if not self.transform:
            return False
        numerical_cols = self.x.select_dtypes(include=[float, int]).columns
        return any(col in numerical_cols for col in self.x.columns)

    def apply(self) ->pandas.DataFrame:
        if not self.check():
            return pandas.concat([self.x, self.y],axis=1)
        self.report_.append("Transformation")
        for col in self.x.select_dtypes(include=[float, int]).columns:
            n = self.x.shape[0]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                stat, p_value_s = shapiro(self.x[col])

            kstest_result = kstest(self.x[col], norm.cdf)
            ks_statistic, p_value_k = kstest_result
            if n > 5000:
                if p_value_k > 0.05:
                    scaler = StandardScaler()
                else:
                    scaler = MinMaxScaler()
            else:
                if p_value_s > 0.05:
                    scaler = StandardScaler()
                else:
                    scaler = MinMaxScaler()
            transformed_col = scaler.fit_transform(self.x[[col]])
            self.x[col] = transformed_col.flatten()
        return pandas.concat([self.x, self.y],axis=1)



import pandas
from mysticML.utils import Daddy

class Encoding(Daddy):
    def init(self, x: pandas.DataFrame, target:str=None,**kwargs) -> None:
        super().__init__(x, target, **kwargs)

    def check(self) ->bool:
        if not self.encode:
            return False
        return any(self.x[col].dtype == object for col in self.x.columns)

    def apply(self) ->pandas.DataFrame:
        if not self.check():
            return pandas.concat([self.x,self.y],axis=1)
        self.report_.append("Encoding")
        for col in self.x.select_dtypes(include=[object]).columns:
            d = dict(self.x[col].value_counts())
            d = {k: v for k, v in sorted(d.items(), key=lambda item: item[1], reverse=True)}
            s = sum(d.values())
            for i,j in d.items():
                d[i] = 1-j/s
            self.x[col] = self.x[col].map(d)
        return pandas.concat([self.x,self.y],axis=1)



import pandas
from scipy.stats import shapiro
from scipy.stats import kstest, norm
from mysticML.utils import Daddy
import warnings

class Imputation(Daddy):
  def __init__(self, x: pandas.DataFrame, target:str=None,**kwargs) -> None:
    super().__init__(x, target, **kwargs)
    self.x = pandas.concat([self.x,self.y],axis=1)

  def check(self) -> bool:
    if not self.impute:
      return False
    si = self.x.shape[0]
    temp = self.x.dropna()
    sf = temp.shape[0]
    if si == sf:
      return False
    else:
      return True

  def apply(self) -> pandas.DataFrame:
    if not self.check():
      return self.x
    self.report_.append("Imputation")
    flag=0
    n = self.x.shape[0]
    null_per = self.x.isnull().mean()
    col_drop = null_per[null_per > 0.6].index
    assert self.target not in col_drop, f"{self.target} is filled mostly with null values ; cannot be a target column."
    self.x.drop(col_drop, axis=1, inplace=True)
    self.x.dropna(subset=[self.target],axis=0,inplace=True)
    rows_with_majority_nan = self.x[(self.x.isna().sum(axis=1) > (self.x.shape[1]/2)).values]
    self.x = self.x.drop(rows_with_majority_nan.index)
    rows_to_drop = self.x[self.x.isna().any(axis=1)].index
    if len(rows_to_drop) <= 0.02 * len(self.x):
      self.x.drop(rows_to_drop, inplace=True)
    for col in self.x.columns:
      if (self.x[col].dtype == 'object') or (self.x[col].dtype == 'int64' and self.x[col].nunique() <= 11 and self.x[col].min() >= 0):
        flag=1
        self.x[col].fillna(self.x[col].mode()[0],inplace=True)
      if self.x[col].dtype in [float, int]:
        t = self.x[col]
        t.dropna(inplace=True)

        with warnings.catch_warnings():
          warnings.simplefilter("ignore")
          stat, p_value_s = shapiro(t)

        kstest_result = kstest(t, norm.cdf)
        ks_statistic, p_value_k = kstest_result
        if n > 5000:
          if p_value_k > 0.05 and flag == 0:
            self.x[col].fillna(self.x[col].mean(), inplace=True)
          else:
            self.x[col].fillna(self.x[col].median(), inplace=True)
        else:
          if p_value_s > 0.05 and flag == 0:
            self.x[col].fillna(self.x[col].mean(), inplace=True)
          else:
            self.x[col].fillna(self.x[col].median(), inplace=True)
    self.x = self.x.reset_index(drop=True)
    return self.x


from mysticML.utils.Daddy import *
from mysticML.utils.Duplicate import *
from mysticML.utils.FeatureExtraction import *
from mysticML.utils.Encoding import *
from mysticML.utils.FeatureSelection import *
from mysticML.utils.Imputation import *
from mysticML.utils.Outlier import *
from mysticML.utils.Sampling import *
from mysticML.utils.Transformation import *


import pandas
from mysticML.utils import Duplicate, Outlier, \
    Imputation, Transformation, Encoding, \
    FeatureSelection, FeatureExtraction, Sampling

class Preprocess(object):
    def __init__(self, df:pandas.DataFrame) ->None:
        '''
            df: accepts pandas.DataFrame
        '''
        self.report_ = []
        self.df = df

    def fit(self, target:str=None, combo:str='advanced',
            duplicate:bool=None, outlier:bool=None, impute:bool=None,
            transform:bool=None, encode:bool=None,
            feature_sel:bool=None, feature_ext:bool=None, sampling:bool=None) ->pandas.DataFrame:
        '''
            target : specify the target column.

            combo : choose from basic | intermediate | advanced

            duplicate : False to force skip, True to include (default)

            outlier : False to force skip, True to include (default)

            impute :  False to force skip, True to include (default)

            transform : False to force skip, True to include (default)

            encode : False to force skip, True to include (default)

            feature_sel : False to force skip, True to include (default)

            feature_ext : False to force skip, True to include (default)

            sampling : False to force skip, True to include (default)
        '''
        df = self.df
        Duplicate.report_.clear()
        operations = [Duplicate, Outlier, Imputation,
        Transformation, Encoding, FeatureSelection, FeatureExtraction, Sampling]

        for operation in operations:
            op = operation(df,target=target,combo=combo,duplicate=duplicate,
                           outlier=outlier,impute=impute,transform=transform,encode=encode,
                           feature_sel=feature_sel,feature_ext=feature_ext,sampling=sampling)
            df = op.apply()
            self.report_ = op.report_

        return df



from mysticML.utils import Daddy
import pandas
pandas.options.mode.chained_assignment=None

class Duplicate(Daddy):
    def __init__(self, x:pandas.DataFrame, target:str=None, **kwargs) ->None:
        super().__init__(x, target, **kwargs)
        self.x = pandas.concat([self.x,self.y],axis=1)

    def check(self) ->bool:
        if not self.duplicate:
            return False
        return True

    def apply(self) ->pandas.DataFrame:
        if not self.check():
            return self.x
        in_size = len(self.x)
        self.x.drop_duplicates(inplace=True)
        if len(self.x) != in_size:
            self.report_.append("Duplicate")
        return self.x